#include "triangle.h"
#include "m3x3.h"
#include "linesegment.h"

#include "scene.h"

Triangle::Triangle(V3 *ovs, V3 *pvs) {

  isReflector = false;
  for (int i = 0; i < 3; i++) {
    overts[i] = ovs[i];
    verts[i] = pvs[i];
  }

}

Triangle::Triangle(V3 v0, V3 v1, V3 v2) {

  isReflector = false;
  verts[0] = v0;
  verts[1] = v1;
  verts[2] = v2;

}

void Triangle::DrawWireframe(FrameBuffer *fb, unsigned int color) {

  for (int i = 0; i < 3; i++) {
    V3 V0 = verts[i];
    V3 V1 = verts[(i+1)%3];
    LineSegment ls(V0, V1);
    ls.Draw2D(fb, color);
  }

}

void Triangle::Rasterize(FrameBuffer *fb, PPC *ppc, 
  unsigned int color, FrameBuffer *text) {

  // Compute and clip bounding box
  float leftf = verts[0][0];
  float rightf = verts[0][0];
  float topf = verts[0][1];
  float bottomf = verts[0][1];

  for (int i = 1; i < 3; i++) {
    if (leftf > verts[i][0])
      leftf = verts[i][0];
    if (rightf < verts[i][0])
      rightf = verts[i][0];
    if (topf > verts[i][1])
      topf = verts[i][1];
    if (bottomf < verts[i][1])
      bottomf = verts[i][1];
  }

  if (bottomf < 0.0f || topf > (float) fb->h || leftf > (float) fb->w ||
    rightf < 0.0f)
    return;

  if (leftf < 0.0f)
    leftf = 0.0f;
  if (rightf > (float)fb->w)
    rightf = (float)fb->w;
  if (topf < 0.0f)
    topf = 0.0f;
  if (bottomf > (float)fb->h)
    bottomf = (float)fb->h;
  // End compute and clip bounding box

  // Setup edge equations
  V3 edges[3];
  for (int i = 0; i < 3; i++) {
    int _i = (i+1)%3;
    edges[i][0] = verts[_i][1]-verts[i][1];
    edges[i][1] = verts[i][0]-verts[_i][0];
    edges[i][2] = -verts[i][0]*edges[i][0]-
      verts[i][1]*edges[i][1];
    if (edges[i]*V3(verts[(i+2)%3][0],  verts[(i+2)%3][1], 1.0f) < 0)
      edges[i] = edges[i]*-1.0f;
  }
  // End setup edge equations

  // Setup rasterization parameter interpolation
  V3 ABCz;
  M3x3 bary;
  bary[0] = V3(verts[0][0], verts[0][1], 1.0f);
  bary[1] = V3(verts[1][0], verts[1][1], 1.0f);
  bary[2] = V3(verts[2][0], verts[2][1], 1.0f);
  bary = bary.Invert();
  ABCz = bary*V3(verts[0][2], verts[1][2], verts[2][2]);
  V3 ABCred;
  ABCred = bary*V3(colors[0][0], colors[1][0], colors[2][0]);
  V3 ABCgreen;
  ABCgreen = bary*V3(colors[0][1], colors[1][1], colors[2][1]);
  V3 ABCblue;
  ABCblue = bary*V3(colors[0][2], colors[1][2], colors[2][2]);

  M3x3 msiQ;
  SetModelSpaceInterpolation(ppc, msiQ);
  V3 rDEFMS; // common denominator for all model space interpolated parameters
  SetModelSpaceInterpolation(msiQ, rDEFMS);
  V3 redABCMS, greenABCMS, blueABCMS;
  SetModelSpaceInterpolation(msiQ, 
    V3(colors[0][0], colors[1][0], colors[2][0]), redABCMS);
  SetModelSpaceInterpolation(msiQ, 
    V3(colors[0][1], colors[1][1], colors[2][1]), greenABCMS);
  SetModelSpaceInterpolation(msiQ, 
    V3(colors[0][2], colors[1][2], colors[2][2]), blueABCMS);
  V3 sABCMS, tABCMS;
  SetModelSpaceInterpolation(msiQ, 
    V3(tcs[0][0], tcs[1][0], tcs[2][0]), sABCMS);
  SetModelSpaceInterpolation(msiQ, 
    V3(tcs[0][1], tcs[1][1], tcs[2][1]), tABCMS);
  V3 normalABCMS[3];
  SetModelSpaceInterpolation(msiQ, 
    V3(normals[0][0], normals[1][0], normals[2][0]), normalABCMS[0]);
  SetModelSpaceInterpolation(msiQ, 
    V3(normals[0][1], normals[1][1], normals[2][1]), normalABCMS[1]);
  SetModelSpaceInterpolation(msiQ, 
    V3(normals[0][2], normals[1][2], normals[2][2]), normalABCMS[2]);

  // End setup rasterization parameter interpolation

  // Rasterization loops: for every pixel inside bounding box
  for (int v = (int) topf; v <= (int) bottomf; v++) {
    for (int u = (int) leftf; u <= (int) rightf; u++) {

      V3 uvf(.5f+(float)u, .5f+(float)v, 1.0f);
      if (uvf*edges[0] < 0.0f 
        || uvf*edges[1] < 0.0f || 
        uvf*edges[2] < 0.0f)
        continue; // outside triangle

      float z = uvf*ABCz;
      if (!fb->Closer(u, v, z))
        continue; // hidden

      // inside the triangle and visible: need to shade
      V3 currColor(uvf*ABCred, uvf*ABCgreen, uvf*ABCblue);
      currColor = V3(uvf*redABCMS, uvf*greenABCMS, 
        uvf*blueABCMS) / (uvf*rDEFMS);
      V3 st = V3(uvf*sABCMS, uvf*tABCMS, 0.0f) / (uvf*rDEFMS);
      if (text) {
        unsigned int uicolor = text->Get((int)(st[0]*(float)text->w), (int)(st[1]*(float)text->h));
        fb->SetGuarded(uvf[0], uvf[1], uicolor, z);
      }
      else {
        fb->SetGuarded(uvf[0], uvf[1], (currColor*255.0f).Color(), z);
      }
      V3 normal = V3(uvf*normalABCMS[0], 
        uvf*normalABCMS[1], uvf*normalABCMS[2]) / (uvf*rDEFMS);
      normal = normal.Normalized();
      V3 P = ppc->GetPoint(u, v, z);
      V3 ray = (ppc->C-P).Normalized();
      V3 rray = normal.Reflect(ray);
      if (isReflector) {
        unsigned int intColor;
        if (scene->impQuad->IntersectBillboard(P, rray, intColor)) {
          V3 rColor(intColor);
          currColor = currColor*255.0f;
          currColor = currColor*0.5f+rColor*0.5f;
          fb->SetGuarded(uvf[0], uvf[1], (currColor).Color(), z);
        }
      }
    }
  }

}

void Triangle::SetModelSpaceInterpolation(PPC *ppc, 
                                          M3x3 &msiQ) {

  msiQ.SetColumn(0, overts[0]-ppc->C);
  msiQ.SetColumn(1, overts[1]-ppc->C);
  msiQ.SetColumn(2, overts[2]-ppc->C);
  M3x3 camM;
  camM.SetColumn(0, ppc->a);
  camM.SetColumn(1, ppc->b);
  camM.SetColumn(2, ppc->c);
  msiQ = msiQ.Invert()*camM;

}

void Triangle::SetModelSpaceInterpolation(M3x3 &msiQ, 
  V3 rvs, V3 &rABCMS) {

  rABCMS[0] = msiQ.GetColumn(0)*rvs;  
  rABCMS[1] = msiQ.GetColumn(1)*rvs;  
  rABCMS[2] = msiQ.GetColumn(2)*rvs;  

}


void Triangle::SetModelSpaceInterpolation(M3x3 &msiQ, 
  V3 &rDEFMS) {

  rDEFMS[0] = msiQ[0][0]+msiQ[1][0]+msiQ[2][0];
  rDEFMS[1] = msiQ[0][1]+msiQ[1][1]+msiQ[2][1];
  rDEFMS[2] = msiQ[0][2]+msiQ[1][2]+msiQ[2][2];

}